from mazemdp.maze import build_maze, create_random_maze
from mazemdp.toolbox import (
    egreedy,
    egreedy_loc,
    random_policy,
    softmax,
    sample_categorical,
)
from ._version import __version__, __version_tuple__



from typing import List, Optional
from pathlib import Path
import anywidget
import ipyreact
from dataclasses import dataclass
import traitlets
import importlib.resources
import numpy as np

from mazemdp.maze import Maze


def get_file(name: str, force_string: bool = False):
    with importlib.resources.path("mazemdp", name) as path:
        if isinstance(path, Path) and not force_string:
            return path
        return path.read_text()


def array_to_json(array: np.ndarray, widget: anywidget.AnyWidget):
    return array.tolist()


@dataclass
class HistoryStep:
    values: np.ndarray
    policy: Optional[np.ndarray]
    agent_state: Optional[int]


class MazeWidget(ipyreact.Widget):
    # Widget front-end JavaScript code
    _esm = get_file("nbwidget.tsx")
    _css = get_file("nbwidget.css", True)

    # Stateful property that can be accessed by JavaScript & Python
    step = traitlets.Int(1).tag(sync=True)

    # Stateful property that can be accessed by JavaScript & Python
    steps = traitlets.Int(1).tag(sync=True)

    # Stateful property that can be accessed by JavaScript & Python
    title = traitlets.Unicode("No title").tag(sync=True)

    # The maze cells
    cells = traitlets.Instance(np.ndarray).tag(sync=True, to_json=array_to_json)
    terminal_states = traitlets.List(traitlets.Int()).tag(sync=True)

    # Values (V or Q, depending on shape)
    policy = traitlets.List(None, allow_none=True).tag(sync=True)

    # Policy (deterministic or not, depending on the shape)
    values = traitlets.List(None, allow_none=True).tag(sync=True)

    # Policy (deterministic or not, depending on the shape)
    agent_state = traitlets.Integer(None, allow_none=True).tag(sync=True)

    #: List of v or q-values
    history: List[HistoryStep]

    def __init__(self, mdp: Maze, title=""):
        super().__init__(
            title="", cells=mdp.cells.T, terminal_states=mdp.terminal_states
        )
        self.mdp = mdp
        self.history = []

    @traitlets.observe("step")
    def observe_step(self, change):
        self.set_trait("values", self.history[change.new - 1].values)
        self.set_trait("policy", self.history[change.new - 1].policy)
        self.set_trait("agent_state", self.history[change.new - 1].agent_state)

    def history_step(self, delta):
        new_step = self.step + delta
        if new_step >= 0 and new_step <= len(self.history):
            self.set_trait("step", new_step)

    def add_step(self, value=None, policy=None, agent_state=None):
        self.history.append(HistoryStep(value, policy, agent_state))

        # Change the # of steps
        self.set_trait("steps", len(self.history))

        # Move to the last
        self.set_trait("step", len(self.history))

    def set_title(self, title: str):
        self.set_trait("title", title)



"""
Author: Olivier Sigaud
"""

from functools import cached_property
import random

import numpy as np

from mazemdp.maze_plotter import MazePlotter  # used to plot the maze
from mazemdp.mdp import Mdp
from mazemdp.toolbox import E, N, S, W


def build_maze(width, height, hit, walls, start_states):
    ts = height * width - 1 - len(walls)
    maze = Maze(
        width,
        height,
        hit=hit,
        walls=walls,
        terminal_states=[ts],
        start_states=start_states,
    )
    return maze.mdp, maze.nb_states, maze.coord_x, maze.coord_y


def build_custom_maze(width, height, hit, walls, terminal_states, start_states):
    maze = Maze(
        width,
        height,
        start_states=start_states,
        walls=walls,
        terminal_states=terminal_states,
        hit=hit,
    )
    return maze.mdp, maze.nb_states, maze.coord_x, maze.coord_y


def create_random_maze(width, height, hit, ratio, start_states):
    size = width * height
    n_walls = round(ratio * size)

    stop = False
    # the loop below is used to check that the maze has a solution
    # if one of the values after check_navigability is null, then another maze should be produced
    while not stop:
        walls = random.sample(range(size), int(n_walls))

        mdp, nb_states, coord_x, coord_y = build_maze(
            width, height, hit, walls, start_states
        )
        stop = mdp.check_navigability()
    return mdp, nb_states, coord_x, coord_y


class Maze:  # describes a maze-like environment
    def __init__(
        self,
        width,
        height,
        hit=False,
        nb_actions=4,
        gamma=0.9,
        timeout=None,
        start_states=None,
        walls=None,
        terminal_states=None,
    ):
        """
        :param width: Int number defining the maze width
        :param height: Int number defining the maze height
        :param walls: List of the states that represent walls in our maze environment
        :param action_list: List of possible actions
        :param nb_actions: used when action_list is empty, by default there are 4 of them (go north, south, eat or west)
        :param gamma: Discount factor of the mdp
        :param timeout: Timeout (deprecated: should be set in gymnasium)
        :param start_states: List defining the states where the agent can be at the beginning of an episode
        :param terminal_states: List defining the states corresponding to the step before the end of an episode
        """
        self.width = width
        self.height = height
        self.cells = np.zeros((width, height), int)
        self.nb_states = height * width - len(walls)

        self.gamma = gamma

        if walls is None:
            walls = []

        self.terminal_states = terminal_states
        if self.terminal_states is None:
            self.terminal_states = []

        self.walls = walls
        self.size = width * height

        self.state_width = []
        self.state_height = []
        self.coord_x = np.zeros(height * width - len(walls) + 1)
        self.coord_y = np.zeros(height * width - len(walls) + 1)
        # ##################### State Space ######################
        self.init_states(width, height, walls)

        # ##################### Action Space ######################
        self.nb_actions = nb_actions

        # ##################### Distribution Over Initial States ######################

        start_distribution = np.zeros(self.nb_states)

        if start_states is None:
            for state in range(len(start_distribution)):
                start_distribution[state] = 1.0 / len(start_distribution)
        else:
            for state in range(len(start_states)):
                start_distribution[state] = 1.0 / len(start_states)

        # ##################### Transition Matrix ######################
        transition_matrix = self.init_transitions(hit)

        if hit:
            reward_matrix = self.reward_hit_walls(transition_matrix)
        else:
            reward_matrix = self.simple_reward(transition_matrix)

        plotter = MazePlotter(self)  # renders the environment

        self.mdp = Mdp(
            self.nb_states,
            self.nb_actions,
            start_distribution,
            transition_matrix,
            reward_matrix,
            plotter,
            gamma=self.gamma,
            terminal_states=self.terminal_states,
            timeout=timeout,
        )

    @cached_property
    def action_space(self):
        """Legacy method to get the action space"""
        import gym

        return gym.spaces.Discrete(self.nb_actions)

    def init_states(self, width, height, walls):
        state = 0
        cell = 0
        for i in range(width):
            for j in range(height):
                if cell not in walls:
                    self.cells[i][j] = state
                    self.coord_x[state] = j
                    self.coord_y[state] = i
                    state = state + 1
                    self.state_width.append(i)
                    self.state_height.append(j)
                else:
                    self.cells[i][j] = -1
                cell = cell + 1

        assert self.nb_states == state, "maze init: error in the number of states"

    def init_transitions(self, hit):
        """
        Init the transition matrix
        """

        transition_matrix = np.empty((self.nb_states, self.nb_actions, self.nb_states))

        transition_matrix[:, N, :] = np.zeros((self.nb_states, self.nb_states))
        transition_matrix[:, S, :] = np.zeros((self.nb_states, self.nb_states))
        transition_matrix[:, E, :] = np.zeros((self.nb_states, self.nb_states))
        transition_matrix[:, W, :] = np.zeros((self.nb_states, self.nb_states))

        for i in range(self.width):
            for j in range(self.height):
                state = self.cells[i][j]
                if not state == -1:

                    # Transition Matrix when going north (no state change if highest cells or cells under a wall)
                    if j == 0 or self.cells[i][j - 1] == -1:
                        transition_matrix[state][N][state] = 1.0
                    else:  # it goes up
                        transition_matrix[state][N][self.cells[i][j - 1]] = 1.0

                    # Transition Matrix when going south (no state change if lowest cells or cells above a wall)
                    if j == self.height - 1 or self.cells[i][j + 1] == -1:
                        transition_matrix[state][S][state] = 1.0
                    else:  # it goes down
                        transition_matrix[state][S][self.cells[i][j + 1]] = 1.0

                    # Transition Matrix when going east (no state change if left cells or on the left side of a wall)
                    if i == self.width - 1 or self.cells[i + 1][j] == -1:
                        transition_matrix[state][E][state] = 1.0
                    else:  # it goes left
                        transition_matrix[state][E][self.cells[i + 1][j]] = 1.0

                    # Transition Matrix when going west (no state change if right cells or on the right side of a wall)
                    if i == 0 or self.cells[i - 1][j] == -1:
                        transition_matrix[state][W][state] = 1.0
                    else:  # it goes right
                        transition_matrix[state][W][self.cells[i - 1][j]] = 1.0

        return transition_matrix

    # --------------------------------- Reward Matrix ---------------------------------
    def simple_reward(self, transition_matrix: np.array):
        reward_matrix = np.zeros((self.nb_states, self.nb_actions))

        for final_state in self.terminal_states:
            for action in range(self.nb_actions):
                reward_matrix[final_state, action] = 1.0

        return reward_matrix

    # --------------------------------- Reward Matrix ---------------------------------
    def reward_hit_walls(self, transition_matrix: np.array):
        # Get the reward for reaching a final state
        reward_matrix = self.simple_reward(transition_matrix)

        # Add negative rewards for hitting a wall
        for i in range(self.width):
            for j in range(self.height):
                state = self.cells[i][j]
                if not state == -1:

                    # Reward Matrix when going north
                    if (
                        j == 0 or self.cells[i][j - 1] == -1
                    ):  # highest cells + cells under a wall
                        reward_matrix[state, N] = -0.5

                    # Reward Matrix when going south
                    if (
                        j == self.height - 1 or self.cells[i][j + 1] == -1
                    ):  # lowest cells + cells above a wall
                        reward_matrix[state, S] = -0.5

                    # Reward Matrix when going east
                    if (
                        i == self.width - 1 or self.cells[i + 1][j] == -1
                    ):  # cells on the left + left side of a wall
                        reward_matrix[state, E] = -0.5

                    # Reward Matrix when going west
                    if (
                        i == 0 or self.cells[i - 1][j] == -1
                    ):  # cells on the right + right side of a wall
                        reward_matrix[state, W] = -0.5

        return reward_matrix


if __name__ == "__main__":
    m = create_random_maze(3, 3, 0.0)



"""
Author: Olivier Sigaud + Antonin Raffin
"""

from dataclasses import dataclass
from typing import Any, List
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import os

try:
    import cv2
except ImportError:
    cv2 = None
import matplotlib

import base64
from pathlib import Path

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.table import Table

from mazemdp.toolbox import E, N, S, W

try:
    shell = get_ipython().__class__.__module__  # noqa: F401
    if shell is not None and shell in ["ipykernel.zmqshell", "google.colab._shell"]:
        os.environ["WEB_NOTEBOOK"] = shell
except NameError:
    pass


# Force backend
if os.environ.get("PLOT_BACKEND"):
    matplotlib.use(os.environ.get("PLOT_BACKEND"))
print(f"Matplotlib backend: {matplotlib.get_backend()}")

# ------------------- plot functions for a maze like environment ----------------#


def show_videos(video_path: str = "", prefix: str = "") -> None:
    """
    Taken from https://github.com/eleurent/highway-env

    :param video_path: Path to the folder containing videos
    :param prefix: Filter the video, showing only the only starting with this prefix
    """
    from IPython import display as ipythondisplay

    html = []

    for avi in Path(video_path).glob(f"{prefix}*.avi"):
        mp4_video = str(avi).replace("avi", "mp4")
        # Convert
        if not os.path.isfile(mp4_video) or os.path.getmtime(
            mp4_video
        ) < os.path.getmtime(avi):
            print(f"Converting {avi}")
            os.system(f"ffmpeg -y -i {avi} -c:v libx264 -crf 19 {mp4_video}")

    for mp4 in Path(video_path).glob(f"{prefix}*.mp4"):
        video_b64 = base64.b64encode(mp4.read_bytes())
        html.append(
            """<video alt="{}" autoplay
                    loop controls style="height: 400px;">
                    <source src="data:video/mp4;base64,{}" type="video/mp4" />
                </video>""".format(
                mp4, video_b64.decode("ascii")
            )
        )
    ipythondisplay.display(ipythondisplay.HTML(data="<br>".join(html)))


def coords(width, height, i, j):
    """
    processes the starting position of the arrows
    i is the width
    j is the height
    """
    x = (0.44 + i) / width
    y = 0.98 - 0.28 / height - j / height
    return x, y


def arrow_params(width, height, i, j, action):
    """
    processes the starting position of the arrows
    """
    x, y = coords(width, height, i, j)

    if action == N:
        return [x, y + 0.08 / height, 0.0, 0.16 / height]
    elif action == S:
        return [x, y - 0.08 / height, 0.0, -0.16 / height]
    elif action == E:
        return [x + 0.16 / width, y, 0.08 / width, 0.0]
    elif action == W:
        return [x - 0.16 / width, y, -0.08 / width, 0.0]
    else:
        return [x, y, 0.0, 0.0]


def qvalue_params(width, height, i, j, action):
    x, y = coords(width, height, i, j)

    if action == N:
        return [x - 0.04 / width, y + 0.32 / height]
    elif action == S:
        return [x - 0.04 / width, y - 0.4 / height]
    elif action == E:
        return [x + 0.32 / width, y]
    elif action == W:
        return [x - 0.24 / width, y]
    else:
        return [x, y]


# -------------------------------------------------------------------------------#


@dataclass
class Plot:
    canvas: FigureCanvasAgg
    fig: Figure
    axes: Axes
    table: Table
    patch: Any


class MazePlotter:
    """
    maze_mdp plot, used to plot the agent in its environment while processing the V/Q function and policy
    it can also create videos given a list of V/Q values and a list of policies
    """

    def __init__(
        self, maze, using_notebook=bool(os.environ.get("WEB_NOTEBOOK", False))
    ):  # maze defined in the mdp notebook
        self.maze_attr = maze
        self.terminal_states = maze.terminal_states
        self.using_notebook = using_notebook
        self.figW = self.maze_attr.width
        self.figH = self.maze_attr.height

        self.plot_history: List[Plot] = []

        self.image_idx = 0
        self.video_writer = None
        self.video_name = ""
        self.video_folder = "videos"

        os.makedirs(self.video_folder, exist_ok=True)

        self.widget_maze = None

    def init_table(self, table: Table):
        """
        the states of the mdp are drawn in a matplotlib table, this function creates this table
        """

        for i in range(self.maze_attr.width):
            for j in range(self.maze_attr.height):
                color = np.zeros(3)
                state = self.maze_attr.cells[i][j]
                if state == -1:
                    color[0] = color[1] = color[2] = 0
                elif state in self.terminal_states:
                    color[0] = 0.9
                    color[1] = 0
                    color[2] = 0.7
                else:
                    color[0] = color[1] = color[2] = 1
                table.add_cell(j, i, 0.1, 0.2, facecolor=color, text="", loc="center")

    def display(self, rgba, mode):
        if mode == "human":
            fig = plt.gcf()
            fig.clear()
            ax = fig.add_subplot()
            ax.axis("off")
            ax.imshow(rgba)
            plt.show(block=False)
        elif mode == "rgb_array":
            return rgba.copy()
        else:
            raise NotImplementedError(f"Mode {mode} is not implemented")

    def render_base(self, title):
        fig = Figure(figsize=(self.figW, self.figH))
        canvas = FigureCanvasAgg(fig)
        axes = fig.add_subplot(111)
        axes.set_title(title)
        table = Table(axes, bbox=[0, 0, 1, 1])
        patch = mpatches.Ellipse(
            (-1, -1), 0.06, 0.085, ec="none", fc="dodgerblue", alpha=0.6
        )

        self.plot_history.append(Plot(canvas, fig, axes, table, patch))
        axes.add_patch(patch)

        self.init_table(table)
        axes.add_table(table)

    def render_notebook(self, title):
        import IPython.display as display
        import ipywidgets as widgets

        # Creates a new widget if needed
        # (1) no widget
        # (2) widget in another cell
        if (
            self.widget_maze is None
            or self.widget_execution_count != get_ipython().execution_count
        ):
            from .nbwidget import MazeWidget

            display.display(display.HTML("<style>" + MazeWidget._css + "</style>"))
            self.widget_maze = MazeWidget(self.maze_attr, title=title)
            play = widgets.Play(
                value=50,
                min=1,
                max=1,
                step=1,
                # Time between two steps in milliseconds
                interval=100,
                description="Press play",
                disabled=False,
            )
            step_backward = widgets.Button(
                description="",
                disabled=False,
                button_style="",
                tooltip="Step backward",
                icon="step-backward",
            )
            step_forward = widgets.Button(
                description="",
                disabled=False,
                button_style="",
                tooltip="Step forward",
                icon="step-forward",
            )
            slider = widgets.IntSlider(min=1, max=1)
            widgets.jslink((play, "value"), (slider, "value"))
            widgets.jslink((slider, "value"), (self.widget_maze, "step"))

            widgets.jslink((self.widget_maze, "steps"), (slider, "max"))
            widgets.jslink((self.widget_maze, "steps"), (play, "max"))

            widget_out = widgets.VBox(
                [
                    self.widget_maze,
                    widgets.HBox([play, step_backward, step_forward]),
                    slider,
                ]
            )
            step_backward.on_click(lambda _: self.widget_maze.history_step(-1))
            step_forward.on_click(lambda _: self.widget_maze.history_step(1))
            self.widget_execution_count = get_ipython().execution_count

            display.display(widget_out)

    def render_notebook_step(self, title, v, policy, agent_state):
        self.render_notebook(title)
        if title:
            self.widget_maze.set_title(title)

        self.widget_maze.add_step(
            value=v.tolist() if v is not None else None,
            policy=policy.tolist() if policy is not None else None,
            agent_state=agent_state,
        )

    def new_render(self, title, mode="human"):
        """
        initializes the plot by creating its basic components (figure, axis,
        agent patch and table) a trace of these components is stored so that the
        old outputs will last on the notebook when a new rendering is performed
        """

        if mode == "human" and self.using_notebook:
            self.render_notebook(title)
        else:
            self.render_base(title)
            canvas = self.plot_history[-1].canvas
            canvas.draw()
            rgba = np.asarray(canvas.buffer_rgba())
            return self.display(rgba, mode)

    def render(
        self,
        v=None,
        policy=None,
        agent_state=None,
        stochastic=False,
        title="No title",
        mode="human",
    ):
        """
        updates the values of the table and the agent position and current
        policy some of these components may not show depending on the parameters
        given when calling this function the agent state is set to None if we do
        not want to plot the agent
        """
        if mode == "human" and self.using_notebook:
            self.render_notebook_step(title, v, policy, agent_state)
            return

        if v is None:
            v = np.array([])

        if policy is None:
            policy = np.array([])

        if len(self.plot_history) == 0:
            self.render_base(title)
        plot = self.plot_history[-1]

        plot.axes.clear()
        plot.axes.add_table(plot.table)

        # Table values and policy update
        for i in range(self.maze_attr.width):
            for j in range(self.maze_attr.height):
                state = self.maze_attr.cells[i][j]
                if len(v) > 0:  # working with state values
                    if len(v.shape) == 1:
                        self.cell_render_v(plot, v, i, j, state)
                    else:  # working with state-action values
                        self.cell_render_q(plot, v, i, j, state)
                if len(policy) > 0:
                    if stochastic:
                        self.render_stochastic_policy(plot, v, policy, i, j, state)
                    else:
                        self.render_policy(plot, policy, i, j, state)

        print(agent_state)

        if agent_state is not None and len(self.maze_attr.state_width) > agent_state:
            x, y = coords(
                self.maze_attr.width,
                self.maze_attr.height,
                self.maze_attr.state_width[agent_state],
                self.maze_attr.state_height[agent_state],
            )
            plot.patch.center = x, y
            radius = 0.1
            circle_patch = mpatches.Circle(
                (x, y), radius=radius, edgecolor="blue", facecolor="blue"
            )
            plot.axes.add_patch(circle_patch)
            plot.axes.add_patch(plot.patch)

        plot.fig.subplots_adjust(left=0.2, bottom=0.2)
        # plot.axes.xticks([])
        # plot.axes.yticks([])
        plot.fig.tight_layout()

        plot.canvas.draw()
        plot.canvas.flush_events()
        rgba = np.asarray(plot.canvas.buffer_rgba())

        return self.display(rgba, mode)

    def cell_render_v(self, plot: Plot, v, i, j, state):
        color = np.zeros(3)
        if state == -1:
            color[0] = color[1] = color[2] = 0
        elif state in self.terminal_states:
            color[0] = 0.9
            color[1] = 0
            color[2] = 0.7
        else:
            color[0] = color[1] = color[2] = np.min([1 - v[state] / (np.max(v) + 1), 1])

        plot.table._cells[(j, i)].set_facecolor(color)
        plot.table._cells[(j, i)]._text.set_text(np.round(v[state], 2))

    def cell_render_q(self, plot: Plot, q, i, j, state):
        color = np.zeros(3)
        if state == -1:
            color[0] = color[1] = color[2] = 0
        elif state in self.terminal_states:
            color[0] = 0.9
            color[1] = 0
            color[2] = 0.7
        else:
            color[0] = color[1] = color[2] = np.min(
                [1 - np.max(q[state]) / (np.max(q) + 1), 1]
            )

        plot.table._cells[(j, i)].set_facecolor(color)
        plot.table._cells[(j, i)]._text.set_text(np.round(np.max(q[state]), 2))

        if not (state == -1 or state in self.terminal_states):
            qmin = np.min(q[state])
            if qmin < 0:
                qmin *= -1  # TODO: should deal better with negative Q values
            pos_q = q[state] + qmin
            qmax = np.max(pos_q)
            norm_q = pos_q / (np.sum(pos_q) - (list(pos_q).count(qmax) * qmax) + 0.1)

            for action in range(len(q[state])):

                x0, y0, x, y = arrow_params(
                    self.maze_attr.width, self.maze_attr.height, i, j, action
                )

                arw_color = "green"
                alpha = 0.9
                qmax = np.max(q[state])

                if not q[state][action] == qmax:
                    arw_color = "red"
                    alpha = norm_q[action]

                if x == 0 and y == 0:
                    circle = mpatches.Circle(
                        (x0, y0), 0.04, ec=arw_color, fc=arw_color, alpha=alpha
                    )
                    plot.axes.add_patch(circle)
                else:
                    plot.axes.arrow(
                        x0,
                        y0,
                        x,
                        y,
                        alpha=alpha,
                        head_width=0.12 / self.maze_attr.width,
                        head_length=0.12 / self.maze_attr.height,
                        fc=arw_color,
                        ec=arw_color,
                    )

    def render_pi(self, policy):
        for i in range(self.maze_attr.width):
            for j in range(self.maze_attr.height):
                state = self.maze_attr.cells[i][j]
                plot = self.plot_history[-1]
                self.render_policy(plot, policy, i, j, state)

    def render_policy(self, plot: Plot, policy, i, j, state):
        if not (state == -1 or state in self.terminal_states):
            x0, y0, x, y = arrow_params(
                self.maze_attr.width, self.maze_attr.height, i, j, policy[state]
            )
            arw_color = "green"
            alpha = 0.6
            plot.axes.arrow(
                x0,
                y0,
                x,
                y,
                alpha=alpha,
                head_width=0.12 / self.maze_attr.width,
                head_length=0.12 / self.maze_attr.height,
                fc=arw_color,
                ec=arw_color,
            )

    def render_stochastic_policy(self, plot: Plot, q, policy, i, j, state):
        color = np.zeros(3)
        if state in self.maze_attr.walls:
            color[0] = color[1] = color[2] = 0
        elif state in self.terminal_states:
            color[0] = 0.9
            color[1] = 0
            color[2] = 0.7
        else:
            color[0] = color[1] = color[2] = np.min(
                [1 - np.max(q[state]) / (np.max(q) + 1), 1]
            )

        plot.table._cells[(j, i)].set_facecolor(color)
        plot.table._cells[(j, i)]._text.set_text(np.round(np.max(q[state]), 2))

        if not (state == -1 or state in self.terminal_states):
            qmin = np.min(q[state])
            if qmin < 0:
                qmin *= -1  # TODO: should deal better with negative Q values
            pos__q = q[state] + qmin
            qmax = np.max(pos__q)
            norm_q = pos__q / (np.sum(pos__q) - (list(pos__q).count(qmax) * qmax) + 0.1)

            for action in range(len(q[state])):
                x0, y0, x, y = arrow_params(
                    self.maze_attr.width, self.maze_attr.height, i, j, action
                )

                q_x, q_y = qvalue_params(
                    self.maze_attr.width, self.maze_attr.height, i, j, action
                )
                arw_color = "green"
                alpha = 0.9
                qmax = np.max(q[state])

                proba = policy[state][action]
                plot.axes.text(q_x, q_y, "{:.2f}".format(proba))

                if not q[state][action] == qmax:
                    arw_color = "red"
                    alpha = norm_q[action]

                if x == 0 and y == 0:
                    circle = mpatches.Circle(
                        (x0, y0), 0.02, ec=arw_color, fc=arw_color, alpha=alpha
                    )
                    plot.axes.add_patch(circle)
                else:
                    plot.axes.arrow(
                        x0,
                        y0,
                        x,
                        y,
                        alpha=alpha,
                        head_width=0.03,
                        head_length=0.03,
                        fc=arw_color,
                        ec=arw_color,
                    )

    def save_fig(self, title):
        self.figure_history[-1].savefig(title)



"""
Author: Olivier Sigaud
"""

from functools import cached_property
import logging
from typing import Optional
import numpy as np
import random

from mazemdp.toolbox import sample_categorical


class Mdp:
    """
    defines a Markov Decision Process
    """

    def __init__(
        self,
        nb_states,
        nb_actions,
        start_distribution,
        transition_matrix,
        reward_matrix,
        plotter,
        gamma=0.9,
        terminal_states=None,
        timeout: Optional[int] = None,
        has_state=True,
    ):
        if timeout is not None:
            assert timeout > 10, "timeout too short:" + timeout
            logging.warning("deprecation: timeout has been set in Mdp (SimpleMazeMDP)")
        self.nb_states = nb_states
        self.nb_actions = nb_actions
        if terminal_states is None:
            terminal_states = []
        self.terminal_states = terminal_states
        self.has_state = has_state
        self.timeout = timeout  # maximum length of an episode
        self.timestep = 0
        self.P0 = start_distribution  # distribution used to draw the first state of the agent, used in method reset()
        self.P = transition_matrix
        self.r = reward_matrix
        self.plotter = plotter  # used to plot the maze
        self.gamma = gamma  # discount factor
        self.current_state = None

    @cached_property
    def action_space(self):
        """Legacy method to get the action space"""
        import gym

        return gym.spaces.Discrete(self.nb_actions)

    def reset(
        self, uniform=False, **kwargs
    ):  # initializes an episode and returns the state of the agent
        # if uniform is set to False, the first state is drawn according to the P0 distribution,
        # else it is drawn from a uniform distribution over all the states except for walls

        if uniform:
            prob = np.ones(self.nb_states) / (self.nb_states)
            self.current_state = sample_categorical(prob)
        else:
            self.current_state = sample_categorical(self.P0)

        self.timestep = 0
        return self.current_state

    def terminated(self):  # returns True if a terminal state was reached
        return self.current_state in self.terminal_states

    def truncated(self):
        """
        truncated=True when timeout reached
        Warning, may collide with the timeLimit when wrapped into a gymnasium environment
        Some the default value of self.timeout is None

        :param self: the MDP
        """
        return self.timestep == self.timeout

    def step(self, u, deviation=0):  # performs a step forward in the environment,
        # if you want to add some noise to the reward, give a value to the deviation param
        # which represents the mean μ of the normal distribution used to draw the noise

        noise = deviation * np.random.randn()  # generate noise, useful for RTDP

        # r is the reward of the transition, you can add some noise to it
        reward = self.r[self.current_state, u] + noise

        # the state reached when performing action u from state x is sampled
        # according to the discrete distribution self.P[x,u,:]
        next_state = sample_categorical(self.P[self.current_state, u, :])

        self.timestep += 1

        info = {
            "State transition probabilities": self.P[self.current_state, u, :],
            "reward's noise value": noise,
        }  # can be used when debugging

        terminated = self.terminated()  # checks if the episode is over
        truncated = self.truncated()  # checks if timit limit was reached
        self.current_state = next_state

        return next_state, reward, terminated, truncated, info

    def sample_transition(self):
        state = random.randint(0, self.nb_states - 1)
        action = random.randint(0, self.nb_actions - 1)
        next_state = sample_categorical(self.P[state, action, :])
        return state, action, next_state

    def new_render(
        self, title, mode="human"
    ):  # initializes a new environment rendering (a plot defined by a figure, an axis...)
        return self.plotter.new_render(title, mode=mode)

    def render(
        self, v=None, policy=None, agent_pos=None, title="No Title", mode="legacy"
    ):  # outputs the agent in the environment with values V (or Q)
        if v is None:
            v = np.array([])

        if policy is None:
            policy = np.array([])

        if not self.has_state:
            return self.plotter.render(v=v, policy=policy, title=title, mode=mode)
        elif agent_pos is not None:
            return self.plotter.render(
                v=v, agent_state=agent_pos, policy=policy, title=title, mode=mode
            )
        elif self.current_state is not None:
            return self.plotter.render(
                v=v,
                agent_state=self.current_state,
                policy=policy,
                title=title,
                mode=mode,
            )
        else:
            return self.plotter.render(v=v, title=title, mode=mode)

    def save_fig(self, title):  # saves the current output into the disk
        self.plotter.save_fig(title)

    def check_navigability(self):
        """
        Checks if any state can be reached from the starting state
        To do so, uses value iteration: any reachable state should get a value above 0
        :param self: the current MDP
        """
        v = np.zeros(self.nb_states)  # initial state values are set to 0
        stop = False

        while not stop:
            v_old = v.copy()

            for x in range(self.nb_states):  # for each state x
                # Compute the value of state x for each action u of the MDP action space
                if x in self.terminal_states:
                    v[x] = np.max(self.r[x, :])
                else:
                    v_temp = []
                    for u in range(self.nb_actions):
                        # Process sum of the values of the neighbouring states
                        summ = 0
                        for y in range(self.nb_states):
                            summ = summ + self.P[x, u, y] * v_old[y]
                        v_temp.append(self.r[x, u] + summ)

                        # Select the highest state value among those computed
                        v[x] = np.max(v_temp)

                        # Test if convergence has been reached
            if (np.linalg.norm(v - v_old)) < 0.01:
                stop = True

        # We should reach terminal states from any starting point
        reachable = self.nb_states - np.count_nonzero(v) == 0
        return reachable



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Author: Olivier Sigaud

Module: Chrono

Description: Class to display time spent in human format rather than seconds
"""
import time


class Chrono:
    def __init__(self):
        self.name = "Chrono"
        self.start = time.time()

    def stop(self):
        stop = time.time()
        dif = stop - self.start
        difstring = ""
        if dif > 3600:
            heures = int(dif / 3600)
            difstring = str(heures) + "h "
            dif = dif - (heures * 3600)
        if dif > 60:
            minutes = int(dif / 60)
            difstring = difstring + str(minutes) + "mn "
            dif = dif - (minutes * 60)
        difstring = difstring + str(int(dif)) + "s "
        dif = int((dif - int(dif)) * 1000)
        difstring = difstring + str(dif) + "ms"
        print(f"Time : {difstring}")
